function [Rz,mu,info] = PSA_subspace_framework(sys,epsilon,bounds,options)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Copyright (C) 2022 Nicat Aliyev, Emre Mengi %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Version 0.1, August 15 2022 
%
%
% call: [Rz,mu,info] = PSA_subspace_framework(sys,epsilon,bounds)
%       [Rz,mu,info] = PSA_subspace_framework(sys,epsilon,bounds,options)
%
%
%
% task: globally optimize a pseudospectral abscissa function
%                       alpha_epsilon(A(mu))
%       for a given 
%           A(mu) =  f_1(mu)*A1 + f_2(mu)*A2 +...+ f_d(mu)*Ad    (1)
%       over mu in a given box (i.e., bounds.lb <= mu <= bounds.ub).
%
%
%
% Input: 
%       sys     A struct containing the information about the system
%       
%       sys.A  (cell of matrices)   
%               an array cell (1xd) consisting of square matrices with
%               the same dimension ordered as in the definition of A(mu),
%               i.e., sys.A{j} = Aj,  where Aj is as in (1)
%
%       sys.f  (cell of function pointers)            
%               function handle specifying the functions in (1)
%               i.e., sys.f(j)=f_j   
%
%       epsilon (real)     
%               (epsilon = 0 implies Rz is spectral abscissa)
%
%		bounds (struct)	 
%               bounds of the box on which the pseudospectral abscissa 
%               is to be optimized
%						   bounds.lb (dx1 real)	 --- lower bounds
%						   bounds.ub (dx1 real)  --- upper bounds
%
%       
%		options (struct)
%
%       options for the *** main subspace framework ***
%
%               options.dimsubspace (integer) -- dim of initial V0  (default=5)
%
%               options.tolpsa (real) --  the tolerance for the accuracy of the computed
%                                         globally optimal value of pseudospectral abscissa (default=10^-7)
%
%               options.mthd (integer) --  1: use eigopt (default for 1 parameter)   
%                                          otherwise: use granso (default for more than one parameter)
%
%               options.maxit (integer) --  maximal # subspace iterations (default=30).   
%
%               options.mu (real)  --  if there is, an estimate for the
%                                      global minimizer of alpha_epsilon(A(x)
%
%               options.print (integer)  --  0: do not print details (default)
%                                            otherwise: print details
%                                                   
%                        
%       options for *** eigopt ***
%
%               options.gamma (real) --  a lower bound for the minimum eigenvalue
%                                        of the Hessian of alpha_epsilon(A^V(mu)) (default = -10)
%					
%               options.itertol(integer)  --  maximal # eigopt iterations (default = 200)
%
%               options.pars (struct)  --  other parameters to be passed to eigopt
% 
%
%       options for *** subspace_pspa ***
%
%               options.maxit_sf  --  maximal # iterations (default 100)
%
%               options.start_z  --  the starting guess (default [], 
%                                 which implies the right-most eigenvalue)
%
%               options.sparse  --  use a sparse EIG and SVD (default issparse(A))
% 
%
%
% Output:
%       Rz     (real)        
%               computed globally smallest value of pseudospectral abscissa
%
%       mu     (dx1 real)    
%              computed global minimizer in the prescribed box
%
%       info   (struct)    
%              consists of additional information about the solution
%
%              info.iter  --  number of subspace iterations until termination
%
%              info.subspace_dim  --  dimension of subspace at termination
%
%              info.Rz_list  --  list of minimal pseudospectral abscissa
%                                for reduced problems 
%
%              info.mu_list  --  list of global minimizers
%                                for reduced problems
%


    
    A=sys.A;
    sf=sys.f;
    
    
    
    
    
    
    %%%%%%%%%%%
    %%%%%%%%%%%
    %%%%% Error Checks first; (E1) - (E4)
    %%%%%%%%%%%
    %%%%%%%%%%%
    
    
    %%%%%%
    % (E1)
    %%%%%%
    if nargin < 4
        options = struct([]);
    end

    if nargin < 3
        error(  'Framework:inputArgumentsMissing',     ...
                'not all necessary input arguments provided.' );
    end 
    %%%%%%%%
    
    
    %%%%%%%
    % (E2)
    %%%%%%%  
    lb = bounds.lb;
    ub = bounds.ub;
    dim = length(lb);
    dim2 = length(ub);
    if (dim ~= dim2)
        error('lengths of lb and ub must be the same');
    end
    %%%%%%%
           
    
    %%%%%%%%
    % (E3)
    %%%%%%%%
    fz = sf(zeros(dim,1));
    if (length(fz) ~= length(A))
        error('lengths of the system matrices and function handles must be the same');
    end    
    %%%%%%%%
    
    
    %%%%%%%
    % (E4)
    %%%%%%%
    
    A1 =  A{1};
    [m,n] = size(A1);
    if (m ~= n)
        error('the matrices A(i) must be squares');
    end
    %%%%%%%%

    
    
    
    
    
    
    
    
    
    %%%%%%%%%%
    %%%%%%%%%%
    %%%% Completing the Missing Options Structure if any; (A1) - (A4)
    %%%%%%%%%%
    %%%%%%%%%%
    
    %%%%%%%
    %% (A1) For the Subspace Framework
    %%%%%%%
    
    if (isfield(options,'tolpsa'))
		tolpsa = options.tolpsa; 
    else
		tolpsa = 10^-7;
    end
    
				
    if (isfield(options,'dimsubspace'))
		dimsubspace = options.dimsubspace;
    else
        dimsubspace = 10;
    end
    
    
    if (isfield(options,'maxit'))
		maxit = options.maxit;
    else
		maxit = 30;
    end
    
    if (isfield(options,'print'))
		printd = options.print;
    else
		printd = 0;
    end
    
    
    if (isfield(options,'mthd'))
		mthd = options.mthd;
    else
        if (dim == 1)
            mthd = 1;            
        else
            mthd = 2;            
        end
    end

    
    if printd
        if (mthd == 1)
            fprintf('the method to be employed for reduced problems is eigopt\n');
        else
            fprintf('the method to be employed for reduced problems is GRANSO\n');
        end        
    end
    
    %%%%%%%
    %%%%%%%
    
    
    
    
    
    
    %%%%%%%%
    %%%%%%%%
    % (A2) For eigopt
    %%%%%%%%
    %%%%%%%%
    
    if (isfield(options,'gamma'))
		eigopt_opts.gamma = options.gamma;
    else
        eigopt_opts.gamma = -10;
    end
    
    eigopt_opts.tol=tolpsa/3;

    if (isfield(options,'itertol'))
		eigopt_opts.itertol = options.itertol;
    else
        eigopt_opts.itertol = 5000;
    end
    
    
    if (isfield(options,'pars'))
        eigopt_opts.pars = options.pars;
    end
    
    eigopt_opts.isprint = 0;
    
    eigopt_opts.Cell_Matrix = A;
    eigopt_opts.Framework_Epsilon = epsilon;
    
    %%%%%%%
    %%%%%%%
    
    
    
   
    
    
    
    %%%%%%%%
    %%%%%%%%
    % (A3) For pspa
    %%%%%%%%
    %%%%%%%%
    
    prtlevel = 0;
    plotfig = 0;
    keyb = 0;
    eigsolver = 0;
    
    %%%%%%%%%
    %%%%%%%%%
    
    
    
    
    
        
    %%%%%%%%%%
    %%%%%%%%%%
    % (A4) For subspace_pspa
    %%%%%%%%%%
    %%%%%%%%%%
    
    if (isfield(options,'maxit'))
		subspace_pspa_opts.maxit = options.maxit;
    else
        subspace_pspa_opts.maxit = 100;
    end
    
    subspace_pspa_opts.ftol = tolpsa/10;

    
    if (isfield(options,'start_z'))
		subspace_pspa_opts.start_z = options.start_z;
    end
    
    if (isfield(options,'sparse'))
		subspace_pspa_opts.sparse = options.sparse;
    end
    
    subspace_pspa_opts.prntlvl = 0; 
    
    %%%%%%%%%
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    
    
   
    
    
    
    
 

    %%%%%%%%%%%%%
    %%%%%%%%%%%%%
    %%%%%%%%%%%%%
    %%%% CONSTRUCT THE INITIAL SUBSPACE
    %%%% V is an orthonormal basis for this subspaces
    %%%%%%%%%%%%%%
    %%%%%%%%%%%%%%
    subspace.V=[];
    zlist = [];
    init_mu_list = [];
    
    
    for i=1:dimsubspace
        
                
        if (dim == 1)

            h=(ub-lb)/(dimsubspace-1);
            mu=lb+(i-1)*h;
            
        else
            
            mu = zeros(dim,1);
            for j=1:dim
                mu(j)=lb(j)+rand()*(ub(j)-lb(j));
            end
            
        end
        
        init_mu_list = [init_mu_list mu];
        
        
        % construct A(mu)
        sd= size(A{1},1);
        M=sparse(sd,sd);
        fk=sf(mu);
        for k=1:length(A)
            M= M+fk(k)*(A{1,k});
        end
        
        
        
        
        
        % find z the rightmost point in the epsln-pseudospec of A(mu)
        if m < 1001
            
        %%% DENSE & SMALL    
                    
            [~, z] = pspa(M, epsilon, prtlevel, plotfig, keyb, eigsolver);
            
            zlist = [zlist z(1,1)];
                        
            % construct the orthonormal basis V
            [~,~,v] = svd(M-z(1,1)*eye(n));
            V=subspace.V;
            subspace.V=orth([V v(:,n)]);
            
        else
            
        %%% SPARSE & LARGE    
        
            [~, infos] = subspace_pspa(M,epsilon,subspace_pspa_opts);
            
            % construct the orthonormal basis V
            if (infos.succ == true)
                z = infos.z_pspa;
                
                zlist = [zlist z(end)];
                
                T = M-z(end)*speye(n);
                [infos.v,~] = eigs(T'*T,1,'smallestabs');
                
                V=subspace.V;
                subspace.V=orth([V infos.v]);
            end
            
        end
        
        if printd
            fprintf('initial subspace construction dim=%d ; completed\n',i);
        end
        
    end
    %%%%%%%%%%%
    %%%%%%%%%%%
    %%%%%%%%%%%
    %%%%%%%%%%%
    

  
        
    %%%%%%%%%%%
    %%%%%%%%%%%
    %%% initializations for the main loop
    %%%%%%%%%%%
    %%%%%%%%%%%
    old_f=0;
    new_f=1000;
    z = -10000;
    
    iter=0;
    subspace.Rz_list=[];
    subspace.mu_list=[];
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    
    
    
    
    
    %%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%
    %%%%% MAIN LOOP
    %%%%%%%%%%%%%%%
    %%%%%%%%%%%%%%%
    
    while (abs(old_f-new_f) >= tolpsa && iter<maxit)        
        
        %%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%
        % (1) Solve the reduced problems
        %%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%                               
        
        
        if mthd ~= 1
            
            %
            % Use GRANSO.
            %
            if printd                
                fprintf('*** Using GRANSO ***\n');
            end
            
            redprob.Current_Subspace = subspace.V;
            redprob.func_handle = sf;
            redprob.Cell_Matrix = sys.A;
            
            for k = 1:length(A)
                redprob.AV{k} = sys.A{k}*subspace.V;
            end
            
            redprob.Framework_Epsilon = epsilon;
            redprob.start_z = zlist;
            redprob.bounds.lb = bounds.lb;
            redprob.bounds.ub = bounds.ub;
            
            objfun = @( mu )feval( 'minimize_pspap', mu, redprob );
            %
            % Construct a function handle for the inequality constraints.
            %
            constrfun = @( mu )feval( 'gransoConstr', mu, redprob );
            %
            % Set GRANSO options.
            %
            
            optgranso.x0 = 0.5*( bounds.lb + bounds.ub );
            optgranso.quadprog_info_msg = false;
            optgranso.opt_tol = tolpsa/3;
            optgranso.print_level = 0;            

            
            solgranso = granso( dim, @( mu )objfun( mu ), ...
                @( mu )constrfun( mu ), [ ], optgranso );

            
            mu = solgranso.final.x;
            Rz = solgranso.final.f;
            
        else
            
            %
            % Use eigopt.
            %
            if printd
                fprintf('*** Using eigopt ***\n');
            end
            
            eigopt_opts.Current_Subspace = subspace.V;
            eigopt_opts.func_handle = sf;
            eigopt_opts.Cell_Matrix = sys.A;
            eigopt_opts.Framework_Epsilon = epsilon;     
            eigopt_opts.start_z = zlist;
            
            for k = 1:length(A)
                eigopt_opts.AV{k} = sys.A{k}*subspace.V;
            end

            
            [Rz,mu,~]=eigopt('minimize_pspap',bounds,eigopt_opts); 
                        
        end
        
        Rz_list = subspace.Rz_list;
        subspace.Rz_list = [Rz_list Rz];
        mu_list = subspace.mu_list;
        subspace.mu_list = [mu_list mu];
        
        
        

        if printd                            
            if isfield(options,'optmu')
                if (dim == 1) 
                    fprintf('iter:%d minimal_pspa: mu=%.12f, |mu-mu*|=%.12f , alpha=%.12f\n', ...
                                iter, mu, abs(mu-options.optmu), Rz);
                else
                    fprintf('iter:%d minimal_pspa: |mu-mu*|=%.12f , alpha=%.12f\n', ...
                                iter, norm(mu-options.optmu), Rz);
                end
            else
                if (dim == 1)
                    fprintf('iter:%d minimal_pspa: mu=%.12f, alpha=%.12f\n', ...
                                iter, mu, Rz);
                else
                    fprintf('iter:%d minimal_pspa: alpha=%.12f\n', ...
                                iter, Rz);
                end
            end
        end
        
        

        % compute A(mu)
        M= sparse(sd,sd);
        fk=sf(mu);
        for k=1:length(A)
            M= M+fk(k)*(A{1,k});
        end
        
        
        
        
        %%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%
        % (2) Expand the subspace, this requires large-scale
        %     computation of pseudospectral abscissa
        %%%%%%%%%%%%%%%%
        %%%%%%%%%%%%%%%%
        
        if m < 1001
            
         %%% DENSE & SMALL 
            %%%%%%%%%%%%
            % (A) if matrices are relatively small, no subspace framework
            %     for pspa computation
            %%%%%%%%%%%%%%%%%%
            
            [~, z] = pspa(M, epsilon, prtlevel, plotfig, keyb, eigsolver);
            
            zlist = [zlist z(1,1)];
            
            [~,~,v] = svd(M-z(1,1)*eye(n));
            V=subspace.V;
            subspace.V=orth([V v(:,n)]);
            
        else
            
         %%% SPARSE & LARGE
            %%%%%%%%%%%%
            % (B) if matrices are really large, subspace framework
            %     is employed
            %%%%%%%%%%%%%%%%%%
            
            [~, infos] = subspace_pspa(M, epsilon,subspace_pspa_opts);
            
            if (infos.succ == true)
                z = infos.z_pspa;
                
                zlist = [zlist z(end)];
                
                T = M-z(end)*speye(n);
                [infos.v,~] = eigs(T'*T,1,'smallestabs');
                
                V=subspace.V;
                subspace.V=orth([V infos.v]);
            end
            
        end
        %%%%%%%%%%%
        %%%%%%%%%%%
        %%%%%%%%%%%
        
        old_f=new_f;
        new_f=Rz;
        iter=iter+1;
        
    end
    
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    % MAIN LOOP ENDS HERE
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    
    
    
    
    
    
    
    %%%%%%%%%%%%
    %%%%%%%%%%%%
    % Set the Output Arguments
    %%%%%%%%%%%%
        
    info.iter = iter;
    [~,info.subspace_dim] = size(subspace.V); 
    info.Rz_list = subspace.Rz_list;
    info.mu_list = subspace.mu_list;

    %%%%%%%%%%%%
    %%%%%%%%%%%%
    %%%%%%%%%%%%


    
return;




