 function [f,fd] = minimize_pspap(mu,opts)
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% Copyright (C) 2022 Nicat Aliyev, Emre Mengi %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Version 0.1, August 15 2022 
% 
%
% Auxiliary Routine; called by eigopt or granso
%
% Output
%      f  : the epsilon pseudospectral abscissa of (A(mu)V-zV) 
%      fd : gradient of the the epsilon pseudospectral abscissa of (A(mu)V-zV)
%
% Note : A(mu) = f1(mu)A1 + ... fd(mu)Ad
%        A(mu)V = f1(mu)A1*V + ... fd(mu)Ad*V
%
%        where
%
%        (1) Aj must be passed through opts.Cell_Matrix{j}
%        (2) Aj*V  through opts.AV
%        (3) fj  through opts.func_handle{j}
%        (4) V  through opts.Current_Subspace
%
%        (5) epsilon must be passed through opts.Framework_Epsilon 
%
% Additional Parameter (not essential to provide) :
%         opts.start_z  --  a list of points possibly contained in the
%                          rightmost component of rectangular pseudospectra





epsilon = opts.Framework_Epsilon;
A = opts.Cell_Matrix;
AV = opts.AV;
sf = opts.func_handle;

V = opts.Current_Subspace;

dim = length(mu);

[~,d]=size(A);
[n1,~]= size(A{1});
[~,r] = size(AV{1});




% Compute A(mu)*V
FV = zeros(n1,r);

fk = sf(mu);
for k = 1:d
    FV = FV + fk(k)*AV{1,k};
end



%%%%%%%%%
%%% Initializations before finding rightmost point
%%% in rectangular pseudospectra
%%%%%%%%%
[~,d2] = size(FV);
R = triu(qr([V FV],0));
Bt = R(1:2*d2,1:d2);
At = R(1:2*d2,d2+1:2*d2);


 bestz = -100;
 
     for Imz = -2:0.1:2

        yB = Imz*Bt;
        
        H1 = [-yB'+sqrt(-1)*At'   epsilon*eye(d2);  ...
                -epsilon*eye(2*d2)      sqrt(-1)*At+yB];
            
        H2 = [Bt'   zeros(d2,d2);   ...
                zeros(2*d2,2*d2)   Bt];
            
        evals = eig(H1,H2);
        
        
        
        [~,inds] = find(abs(real(evals)) < 10^-8);
        
        
        if ~isempty(inds)            
            mRz = max(imag(evals(inds)));
                            
            if mRz > real(bestz)
                bestz = mRz + sqrt(-1)*Imz;
            end        
            
        end
        
    end
  
if bestz < -99.99
    pencil_pspa_opts.start_z = opts.start_z;
else
    pencil_pspa_opts.start_z = [opts.start_z bestz];
end
%%%%%%%%%%%%
%%%%%%%%%%%%



%%%%%%%%
%%%%%%%%
%%% Compute the function value
%%%%%%%%
[f, z, ~, ~] = pencil_pspa(At, Bt, epsilon,pencil_pspa_opts);
%%%%%
%%%%%





%%%%%%
%%% Derivatives of fk at mu for k = 1,...,d 
%%%%%%%

    h=10^(-8);
    
    fc = sf(mu);
    
    for j = 1:dim
        
        mup = mu;
        mup(j) = mup(j) + h;
        fp = sf(mup);
            
        df(:,j) = ( fp - fc )/h;
        
    end


 

%%%%%%%%%%
%%% Derivative of A(mu)
%%%%%%%%%%%

for j = 1:dim

    
    dAV{j} = zeros(n1,r);
    
    for k = 1:length(A)
        dAV{j} = dAV{j} + df(k,j)*AV{1,k};
    end
    
end



%%%%%%%%
% Calculate the derivative of the epsilon-PSA for reduced problems 
%%%%%%%%

[Ud,~,Vd] = svd(FV-z(1)*V,'econ');
u = Ud(:,end);
v = Vd(:,end);

% partial derivatives of the epsilon-PSA for reduced problems 
for j = 1:dim
    
    fd(j,1) = real ( (u'*dAV{j}*v)/(u'*V*v) );
    
end

   
end

